
/* ******************************************************************
   FSE : Finite State Entropy coder
   header file
   Copyright (C) 2013-2014, Yann Collet.
   BSD 2-Clause License (http://www.opensource.org/licenses/bsd-license.php)

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
       * Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the
   distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   You can contact the author at :
   - Public forum : https://groups.google.com/forum/#!forum/lz4c
****************************************************************** */
#pragma once

#if defined (__cplusplus)
extern "C" {
#endif


/******************************************
   Compiler Options
******************************************/
#if defined(_MSC_VER) && !defined(__cplusplus)   // Visual Studio
#  define inline __inline           // Visual C is not C99, but supports some kind of inline
#endif


/******************************************
   Includes
******************************************/
#include <stddef.h>    // size_t, ptrdiff_t


/******************************************
   FSE simple functions
******************************************/
int FSE_compress   (void* dest,const unsigned char* source, int sourceSize);
int FSE_decompress (unsigned char* dest, int originalSize,const void* compressed);
/*
FSE_compress():
    Compress table of unsigned char 'source', of size 'sourceSize', into destination buffer 'dest'.
    'dest' buffer must be already allocated, and sized to handle worst case situations.
    Use FSE_compressBound() to determine this size.
    return : size of compressed data
             or -1 if there is an error.
FSE_decompress():
    Decompress compressed data from buffer 'compressed',
    into destination table of unsigned char 'dest', of size 'originalSize'.
    Destination table must be already allocated, and large enough to accomodate 'originalSize' char.
    The function will determine how many bytes are read from buffer 'compressed'.
    return : size of compressed data
             or -1 if there is an error.
*/
#define FSE_MAX_HEADERSIZE 512
#define FSE_COMPRESSBOUND(size) (size + FSE_MAX_HEADERSIZE)
static inline int FSE_compressBound(int size) { return FSE_COMPRESSBOUND(size); }
/*
FSE_compressBound():
    Gives the maximum (worst case) size that can be reached by function FSE_compress.
    Used to know how much memory to allocate for destination buffer.
*/


/******************************************
   FSE advanced functions
******************************************/
/*
FSE_compress2():
    Same as FSE_compress(), but allows the selection of 'nbSymbols' and 'tableLog'
    Both parameters can be defined as '0' to mean : use default value
    The function will then assume that any unsigned char within 'source' has value < nbSymbols.
    note : If this condition is not respected, compressed data will be corrupted !
    return : size of compressed data
             or -1 if there is an error
*/
int FSE_compress2 (void* dest, const unsigned char* source, int sourceSize, int nbSymbols, int tableLog);


/*
FSE_decompress_safe():
    Same as FSE_decompress(), but ensures that the decoder never reads beyond compressed + maxCompressedSize.
    note : you don't have to provide the exact compressed size. If you provide more, it's fine too.
    This function is safe against malicious data.
    return : size of compressed data
             or -1 if there is an error
*/
int FSE_decompress_safe (unsigned char* dest, int originalSize, const void* compressed, int maxCompressedSize);


/******************************************
   FSE detailed API
******************************************/
/*
int FSE_compress(char* dest, const char* source, int inputSize) does the following:
1. count symbol occurence from table source[] into table count[]
2. normalize counters so that sum(counting[]) == Power_of_2 (== 2^tableLog)
3. saves normalized counters to memory buffer using writeHeader()
4. builds encoding tables from the normalized counters
5. encodes the data stream using encoding tables

int FSE_decompress(char* dest, int originalSize, const char* compressed) performs:
1. reads normalized counters with readHeader()
2. builds decoding tables from the normalized counters
3. decodes the data stream using these decoding tables

The following API allows to target specific sub-functions.
*/

/* *** COMPRESSION *** */

int FSE_count(unsigned int* count, const unsigned char* source, int sourceSize, int maxNbSymbols);

int FSE_normalizeCount(short* normalizedCounter, int maxTableLog, unsigned int* count, int total, int nbSymbols);

static inline int FSE_headerBound(int nbSymbols, int tableLog) { (void)tableLog; return nbSymbols ? (nbSymbols*2)+1 : 512; }
int FSE_writeHeader (void* header, const short* normalizedCounter, int nbSymbols, int tableLog);

int FSE_sizeof_CTable(int nbSymbols, int tableLog);
int FSE_buildCTable(void* CTable, const short* normalizedCounter, int nbSymbols, int tableLog);

int FSE_compress_usingCTable (void* dest, const unsigned char* source, int sourceSize, const void* CTable);

/*
The first step is to count all symbols. FSE_count() provides one quick way to do this job.
Result will be saved into 'count', a table of unsigned int, which must be already allocated, and have 'maxNbSymbols' cells.
'source' is assumed to be a table of char of size 'sourceSize' if 'maxNbSymbols' <= 256.
All values within 'source' MUST be < maxNbSymbols.
FSE_count() will return the highest symbol value detected into 'source' (necessarily <= 'maxNbSymbols', can be 0 if only 0 is present).
If there is an error, the function will return -1.

The next step is to normalize the frequencies, so that Sum_of_Frequencies == 2^tableLog.
This is performed by function FSE_normalizeCount()
Alternatively, function FSE_normalizeCountHC() do the same, but is slower and provides the best achievable normalization.
The result will be saved into a structure, called 'normalizedCounter', which is a table of unsigned int.
'normalizedCounter' must be already allocated, and have 'nbSymbols' cells.
FSE_normalizeCount() will ensure that sum of 'nbSymbols' frequencies is == 2 ^'tableLog', it also guarantees a minimum of 1 to any Symbol which frequency is >= 1.
FSE_normalizeCount() can work "in place" to preserve memory, using 'count' as both source and destination area.
You can use input 'tableLog'==0 to mean "default value".
The return value is the adjusted tableLog. It is necessary to retrieve it for next steps.
A result of '0' means that there is only a single symbol present.
If there is an error, the function will return -1.

'normalizedCounter' can be saved in a compact manner to a memory area using FSE_writeHeader().
The target memory area must be pointed by 'header'.
'header' buffer must be already allocated. Its size must be at least FSE_headerBound().
The result of the function is the number of bytes written into 'header'.
If there is an error, the function will return -1.

'normalizedCounter' can then be used to create the compression tables 'CTable'.
The space required by 'CTable' must be already allocated. Its size is provided by FSE_sizeof_CTable().
You can then use FSE_buildCTable() to fill 'CTable'.
In both cases, if there is an error, the function will return -1.

'CTable' can then be used to compress 'source', with FSE_compress_usingCTable().
Similar to FSE_count(), the convention is that 'source' is assumed to be a table of char of size 'sourceSize'
The function returns the size of compressed data (without header), or -1 if failed.
*/


/* *** DECOMPRESSION *** */

int FSE_readHeader (short* const normalizedCounter, int* nbSymbols, int* tableLog, const void* header);

int FSE_sizeof_DTable(int tableLog);
int FSE_buildDTable (void* DTable, const short* const normalizedCounter, int nbSymbols, int tableLog);

int FSE_decompress_usingDTable(unsigned char* dest, const int originalSize, const void* compressed, const void* DTable, const int tableLog);

/*
The first step is to get the normalized frequency of symbols.
This can be performed by reading a header with FSE_readHeader().
'normalizedCounter' must be already allocated, and have at least 'nbSymbols' cells.
In practice, that means it's necessary to know 'nbSymbols' beforehand,
or size it to handle worst case situations (typically 256).
FSE_readHeader will provide 'tableLog' and 'nbSymbols' stored into the header.
The result of FSE_readHeader() is the number of bytes read from 'header'.
The following values have special meaning :
return 2 : there is only a single symbol value. The value is provided into the second byte.
return 1 : data is uncompressed
If there is an error, the function will return -1.

The next step is to create the decompression tables 'DTable' from 'normalizedCounter'.
This is performed by the function FSE_buildDTable().
The space required by 'DTable' must be already allocated. Its size is provided by FSE_sizeof_DTable().

'DTable' can then be used to decompress 'compressed', with FSE_decompress_usingDTable().
FSE_decompress_usingDTable() will regenerate exactly 'originalSize' symbols, as a table of unsigned char.
The function returns the size of compressed data (without header), or -1 if failed.
*/


/******************************************
   FSE streaming API
******************************************/
typedef struct
{
    size_t bitContainer;
    int bitPos;
} bitStream_forward_t;

void* FSE_initCompressionStream(void** op);
void FSE_initStateAndPtrs(ptrdiff_t* state, const void** CTablePtr1, const void** CTablePtr2, const void* CTable);
void FSE_encodeByte(ptrdiff_t* state, bitStream_forward_t* bitC, unsigned char symbol, const void* CTablePtr1, const void* CTablePtr2);
static void FSE_addBits(bitStream_forward_t* bitC, size_t value, int nbBits);
static void FSE_flushBits(void** outPtr, bitStream_forward_t* bitC);
int FSE_closeCompressionStream(void* outPtr, bitStream_forward_t* bitC, void* compressionStreamDescriptor, int optionalId);

/*
These function allow the creation of custom streams, mixing mutiple tables and bit sources.
They are used by FSE_compress_usingCTable().

A key property to keep in mind is that encoding and decoding are done **in reverse direction**.
So the first symbol you will encode is the last you will decode, like a lifo stack.
This logic applies to any bitstream value inserted into the bitstream.

You will need a few variables to track your bitStream. They are :

void* op;           // Your output buffer (must be already allocated)
void* compressionStreamDescriptor;   // Required to init and close the bitStream
void* CTable;       // Provided by FSE_buildCTable()
ptrdiff_t state;    // Encode fractional bits
bitStream_forward_t bitStream={0,0}; // Store bitStream into register before writing it
void* CTablePtr1;   // Encoding Table n1. Provided by init. Required by encodeByte.
void* CTablePtr2;   // Encoding Table n2. Provided by init. Required by encodeByte.


The first thing to do is to init the bitStream.
    void* compressionStreamDescriptor = FSE_initCompressionStream(&op);

And then init your state and Ptrs. Ptrs are required by the encoded function.
    FSE_initStateAndPtrs(&state, &CTablePtr1, &CTablePtr2, CTable);

You can then encode your input data, byte after byte.
You are free to choose the direction in which you encode, as long as you remember decoding will be done in reverse direction.
    FSE_encodeByte(&state, &bitStream, symbol, CTablePtr1, CTablePtr2);

At any time, you can add any other bit sequence.
Note : maximum allowed nbBits is 25, to be compatible with 32-bits decoders
    FSE_addBits(&bitStream, bitField, nbBits);

Writing data to memory is performed by the flush method.
It's possible to store several bitFields into bitStream before calling flush a single time.
BitStream size is 64-bits on 64-bits systems, 32-bits on 32-bits systems (size_t).
The nb of bits already written into bitStream is stored into bitPos.
For information, FSE_encodeByte() never writes more than 'tableLog' bits at a time.
    FSE_flushBits(&op, &bitStream);

Your last FSE encoding operation shall be to flush your state value.
    FSE_addBits(&bitStream, state, tableLog);
    FSE_flushBits(&op, &bitStream);

When you are done with compression, you must close the bitStream.
It's possible to embed an optionalId into the header, for later information, value must be between 1 and 4.
The function returns the size in bytes of the compressed stream.
If there is an error, it returns -1.
    int size = FSE_closeCompressionStream(op, &bitStream, compressionStreamDescriptor, optionalId);
*/

typedef struct
{
    unsigned int bitContainer;
    int bitsConsumed;
} bitStream_backward_t;

const void* FSE_initDecompressionStream (const void** p, bitStream_backward_t* bitC, int* optionalId);
const void* FSE_initDecompressionStream_safe (const void** p, bitStream_backward_t* bitC, int* optionalId, int maxCompressedSize);
unsigned char FSE_decodeSymbol(unsigned int* state, bitStream_backward_t* bitC, const void* DTable);
unsigned int FSE_readBits(bitStream_backward_t* bitC, unsigned int nbBits);
void FSE_updateBitStream(bitStream_backward_t* bitC, const void** ip);
int FSE_closeDecompressionStream(const void* decompressionStreamDescriptor, const void* input);

/*
Now is the turn to decompose FSE_decompress_usingDTable().
You will decode FSE_encoded symbols from the bitStream,
but also any other bitFields you put in, **in reverse order**.
So, typically, if you encoded from end to beginning, you will now decode from beginning to end.

You will need a few variables to track your bitStream. They are :

const void* input;        // Your input buffer (where compressed data is)
const void* decompressionStreamDescriptor;   // Required to init and close the bitStream
const void* DTable;       // Provided by FSE_buildDTable()
int   tableLog;           // Provided by FSE_readHeader()
unsigned int state;       // Encoded fractional bits
bitStream_backward_t bitStream;   // Store bits read from input

The first thing to do is to init the bitStream.
    decompressionStreamDescriptor = FSE_initDecompressionStream(&ip, &bitStream, &optionalId);

You should then retrieve your initial state value :
    state = FSE_readBits(&bitC, tableLog);
    FSE_updateBitStream(&bitStream, &ip);

You can then decode your data, byte after byte.
Keep in mind data is decoded in reverse order, like a lifo container.
    unsigned char symbol = FSE_decodeSymbol(&state, bitStream, &bitsConsumed, DTable);

You can retrieve any other bitfield you eventually stored into the bitStream (in reverse order)
Note : maximum allowed nbBits is 25
    unsigned int bitField = FSE_readBits(&bitStream, nbBits);

Reading data to memory is performed by the update method.
'bitConsumed' shall never > 32.
For information the maximum number of bits read by FSE_decodeSymbol() is 'tableLog'.
Don't hesitate to use this method. The cost of this operation is very low.
    FSE_updateBitStream(&bitStream, &ip);

When you are done with decompression, you can close the bitStream (optional).
The function returns the size in bytes of the compressed stream.
If there is an error, it returns -1.
    int size = FSE_closeDecompressionStream(decompressionStreamDescriptor, ip, bitsConsumed);
*/


/***********************************************************************
   *** inlined functions (for performance) ***
   GCC is not as good as visual to inline code from other *.c files
***********************************************************************/

static inline void FSE_addBits(bitStream_forward_t* bitC, size_t value, int nbBits)
{
    static const unsigned int mask[] = { 0, 1, 3, 7, 0xF, 0x1F, 0x3F, 0x7F, 0xFF, 0x1FF, 0x3FF, 0x7FF, 0xFFF, 0x1FFF, 0x3FFF, 0x7FFF, 0xFFFF, 0x1FFFF, 0x3FFFF, 0x7FFFF, 0xFFFFF, 0x1FFFFF, 0x3FFFFF, 0x7FFFFF,  0xFFFFFF, 0x1FFFFFF };   // up to 25 bits
    bitC->bitContainer |= (value & mask[nbBits]) << bitC->bitPos;
    bitC->bitPos += nbBits;
}

static inline void FSE_flushBits(void** outPtr, bitStream_forward_t* bitC)
{
    ** (size_t**) outPtr = bitC->bitContainer;
    {
        size_t nbBytes = bitC->bitPos >> 3;
        bitC->bitPos &= 7;
        *(char**)outPtr += nbBytes;
        bitC->bitContainer >>= nbBytes*8;
    }
}

///  F S E - ENCODER ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#if defined (__cplusplus)
}
#endif
/* ******************************************************************
   FSE : Finite State Entropy coder
   Copyright (C) 2013-2014, Yann Collet.
   BSD 2-Clause License (http://www.opensource.org/licenses/bsd-license.php)

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

       * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
       * Redistributions in binary form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the
   distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   You can contact the author at :
   - Public forum : https://groups.google.com/forum/#!forum/lz4c
****************************************************************** */

#ifndef FSE_DONTINCLUDECORE

//****************************************************************
// Tuning parameters
//****************************************************************
// MEMORY_USAGE :
// Memory usage formula : N->2^N Bytes (examples : 10 -> 1KB; 12 -> 4KB ; 16 -> 64KB; 20 -> 1MB; etc.)
// Increasing memory usage improves compression ratio
// Reduced memory usage can improve speed, due to cache effect
// Recommended max value is 14, for 16KB, which nicely fits into Intel x86 L1 cache
#define FSE_MAX_MEMORY_USAGE 14
#define FSE_DEFAULT_MEMORY_USAGE 13

// FSE_MAX_NB_SYMBOLS :
// Maximum nb of symbol values authorized.
// Required for proper stack allocation
#define FSE_MAX_NB_SYMBOLS 256

// FSE_ILP :
// Determine if the algorithm tries to explicitly exploit ILP
// (Instruction Level Parallelism)
// Default : Recommended
#define FSE_ILP 1


/****************************************************************
   Generic function type & suffix (C template temulation)
****************************************************************/
#define FSE_FUNCTION_TYPE BYTE
#define FSE_FUNCTION_EXTENSION

#endif    // FSE_DONTINCLUDECORE


//****************************************************************
//* Includes
//****************************************************************

#include <stddef.h>    // ptrdiff_t
#include <string.h>    // memcpy, memset
#include <stdio.h>     // printf (debug)


//****************************************************************
//* Basic Types
//****************************************************************
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L   // C99
# include <stdint.h>
typedef  uint8_t BYTE;
typedef uint16_t U16;
typedef  int16_t S16;

typedef  int32_t S32;
typedef u__int64 U64;
typedef  __int64 S64;
#else
typedef unsigned char       BYTE;
typedef unsigned short      U16;
typedef   signed short      S16;

typedef   signed int        S32;
typedef unsigned long long  U64;
typedef   signed long long  S64;
#endif

typedef size_t bitContainer_t;
typedef size_t scale_t;


//****************************************************************
//* Constants
//****************************************************************
#define FSE_MAX_TABLELOG  (FSE_MAX_MEMORY_USAGE-2)
#define FSE_MAX_TABLESIZE (1U<<FSE_MAX_TABLELOG)
#define FSE_MAXTABLESIZE_MASK (FSE_MAX_TABLESIZE-1)
#define FSE_DEFAULT_TABLELOG (FSE_DEFAULT_MEMORY_USAGE-2)
#define FSE_MIN_TABLELOG 5

#define FSE_VIRTUAL_LOG   ((sizeof(scale_t)*8)-2)
#define FSE_VIRTUAL_RANGE ((scale_t)1<<FSE_VIRTUAL_LOG)

#if FSE_MAX_TABLELOG>15
#error "FSE_MAX_TABLELOG>15 isn't supported"
#endif


//****************************************************************
//* Compiler specifics
//****************************************************************
#ifdef _MSC_VER    // Visual Studio
#  define FORCE_INLINE static __forceinline
#  include <intrin.h>                    // For Visual 2005
#  pragma warning(disable : 4127)        // disable: C4127: conditional expression is constant
#  pragma warning(disable : 4214)        // disable: C4214: non-int bitfields
#else
#  define GCC_VERSION (__GNUC__ * 100 + __GNUC_MINOR__)
#  ifdef __GNUC__
#    define FORCE_INLINE static inline __attribute__((always_inline))
#  else
#    define FORCE_INLINE static inline
#  endif
#endif


/****************************************************************
  Complex types
****************************************************************/
typedef struct
{
    int  deltaFindState;
    U16  maxState;
    unsigned char minBitsOut;
} FSE_symbolCompressionTransform;

typedef struct
{
    U16 tableLog;
    U16 nbSymbols;
    U16 stateTable[FSE_MAX_TABLESIZE];
    FSE_symbolCompressionTransform symbolTT[FSE_MAX_NB_SYMBOLS];   // Also used by FSE_compressU16
} CTable_max_t;


/****************************************************************
  Internal functions
****************************************************************/
FORCE_INLINE int FSE_highbit (register U32 val)
{
#   if defined(_MSC_VER)   // Visual
    unsigned long r;
    _BitScanReverse ( &r, val );
    return (int) r;
#   elif defined(__GNUC__) && (GCC_VERSION >= 304)   // GCC Intrinsic
    return 31 - __builtin_clz (val);
#   else   // Software version
    static const int DeBruijnClz[32] = { 0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30, 8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31 };
    U32 v = val;
    int r;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    r = DeBruijnClz[ (U32) (v * 0x07C4ACDDU) >> 27];
    return r;
#   endif
}

static short FSE_abs(short a) { return a<0? -a : a; }


#ifndef FSE_DONTINCLUDECORE

/****************************************************************
   Header bitstream management
****************************************************************/
int FSE_writeHeader (void* header, const short* normalizedCounter, int nbSymbols, int tableLog)
{
    BYTE* const ostart = (BYTE*) header;
    BYTE* out = ostart;
    int nbBits;
    const int tableSize = 1 << tableLog;
    int remaining;
    int threshold;
    U32 bitStream;
    int bitCount;
    int charnum = 0;
    int previous0 = 0;

    if (tableLog > FSE_MAX_TABLELOG) return -1;   // Unsupported
    if (tableLog < FSE_MIN_TABLELOG) return -1;   // Unsupported

    // HeaderId (normal case)
    bitStream = 2;
    bitCount  = 2;
    // Table Size
    bitStream += (tableLog-FSE_MIN_TABLELOG) << bitCount;
    bitCount  += 4;

    // Init
    remaining = tableSize+1;   // +1 for extra accuracy
    threshold = tableSize;
    nbBits = tableLog+1;

    while (remaining>1)   // stops at 1
    {
        if (previous0)
        {
            int start = charnum;
            while (!normalizedCounter[charnum]) charnum++;
            while (charnum >= start+24) { start+=24; bitStream += 0xFFFF<<bitCount; *(U16*)out=(U16)bitStream; out+=2; bitStream>>=16; }
            while (charnum >= start+3) { start+=3; bitStream += 3 << bitCount; bitCount += 2; }
            bitStream += (charnum-start) << bitCount; bitCount += 2;
            if (bitCount>16)
            {
                *(U16*)out = (U16)bitStream;
                out += 2;
                bitStream >>= 16;
                bitCount -= 16;
            }
        }
        {
            short count = normalizedCounter[charnum++];
            const short max = (short)((2*threshold-1)-remaining);
            remaining -= FSE_abs(count);
            count++;   // +1 for extra accuracy
            if (count>=threshold) count += max;   // [0..max[ [max..threshold[ (...) [threshold+max 2*threshold[
            bitStream += count << bitCount;
            bitCount  += nbBits;
            bitCount  -= (count<max);
            previous0 = (count==1);
            while (remaining<threshold) { nbBits--; threshold>>=1; }
        }
        if (bitCount>16)
        {
            *(U16*)out = (U16)bitStream;
            out += 2;
            bitStream >>= 16;
            bitCount -= 16;
        }
    }

    if (remaining<0) return -1;

    * (U16*) out = (U16) bitStream;
    out+= (bitCount+7) /8;

    if (charnum > nbSymbols) return -1;   // Too many symbols written

    return (int) (out-ostart);
}


int FSE_readHeader (short* const normalizedCounter, int* nbSymbols, int* tableLog, const void* header)
{
    const BYTE* const istart = (const BYTE*) header;
    const BYTE* ip = (const BYTE*) header;
    int nbBits;
    int remaining;
    int threshold;
    U32 bitStream;
    int bitCount;
    int charnum = 0;
    int previous0 = 0;

    bitStream = * (U32*) ip;
    bitStream >>= 2;
    nbBits = (bitStream & 0xF) + FSE_MIN_TABLELOG;   // read tableLog
    bitStream >>= 4;
    *tableLog = nbBits;
    remaining = (1<<nbBits)+1;
    threshold = 1<<nbBits;
    nbBits++;
    bitCount = 6;

    while (remaining>1)
    {
        if (previous0)
        {
            int n0 = charnum;
            while ((bitStream & 0xFFFF) == 0xFFFF) { n0+=24; ip+=2; bitStream = (*(U32*)ip) >> bitCount; }
            while ((bitStream & 3) == 3) { n0+=3; bitStream>>=2; bitCount+=2; }
            n0 += bitStream & 3; bitCount += 2;
            while (charnum < n0) normalizedCounter[charnum++] = 0;
            ip += bitCount>>3; bitCount &= 7; bitStream = (*(U32*)ip) >> bitCount;
        }
        {
            const short max = (short)((2*threshold-1)-remaining);
            short count;

            if ((bitStream & (threshold-1)) < (U32)max)
            {
                count = (short)(bitStream & (threshold-1));
                bitCount   += nbBits-1;
            }
            else
            {
                count = (short)(bitStream & (2*threshold-1));
                if (count >= threshold) count -= max;
                bitCount   += nbBits;
            }

            count--;   // extra accuracy
            remaining -= FSE_abs(count);
            normalizedCounter[charnum++] = count;
            previous0 = !count;
            while (remaining < threshold) { nbBits--; threshold >>= 1; }

            ip += bitCount>>3; bitCount &= 7; bitStream = (*(U32*)ip) >> bitCount;
        }
    }
    *nbSymbols = charnum;
    if (remaining < 1) return -1;
    if (nbBits > FSE_MAX_TABLELOG) return -1;  // Too large

    ip += bitCount>0;
    return (int) (ip-istart);
}


//****************************
// FSE Compression Code
//****************************

/*
CTable is a variable size structure which contains :
    U16 tableLog;
    U16 nbSymbols;
    U16 nextStateNumber[1 << tableLog];                   // This size is variable
    FSE_symbolCompressionTransform symbolTT[nbSymbols];   // This size is variable
Allocation is manual, since C standard does not support variable-size structures.
*/
#define FSE_SIZEOF_CTABLE_U32(s,t) (((((2 + (1<<t))*sizeof(U16)) + ((s+1)*sizeof(FSE_symbolCompressionTransform)))+(sizeof(U32)-1)) / sizeof(U32))
int FSE_sizeof_CTable (int nbSymbols, int tableLog)
{
    if (tableLog > FSE_MAX_TABLELOG) return -1;   // Max supported value
    return (int) (FSE_SIZEOF_CTABLE_U32 (nbSymbols, tableLog) * sizeof (U32) );
}

typedef struct
{
    U16  newState;
    BYTE symbol;
    BYTE nbBits;
} FSE_decode_t;


// Emergency distribution strategy (fallback of fallback); compression will be seriously hurt ; consider increasing table size
static void FSE_emergencyDistrib(short* normalizedCounter, int nbSymbols, short points)
{
    int s=0;
    while (points)
    {
        if (normalizedCounter[s] > 1)
        {
            normalizedCounter[s]--;
            points--;
        }
        s = (s+1) % nbSymbols;
    }
}

// fallback distribution (corner case); compression will be hurt ; consider increasing table size
static void FSE_distribNpts(short* normalizedCounter, int nbSymbols, short points)
{
    int s;
    int rank[5] = {0};
    int fallback=0;

    // Sort 4 largest (they'll absorb normalization rounding)
    for (s=1; s<nbSymbols; s++)
    {
        int i, b=3;
        if (b>=s) b=s-1;
        while ((b>=0) && (normalizedCounter[s]>normalizedCounter[rank[b]])) b--;
        for (i=3; i>b; i--) rank[i+1] = rank[i];
        rank[b+1]=s;
    }

    // Distribute points
    s = 0;
    while (points)
    {
        short limit = normalizedCounter[rank[s+1]]+1;
        if (normalizedCounter[rank[s]] >= limit + points )
        {
            normalizedCounter[rank[s]] -= points;
            break;
        }
        points -= normalizedCounter[rank[s]] - limit;
        normalizedCounter[rank[s]] = limit;
        s++;
        if (s==3)
        {
            short reduction = points>>2;
            if (fallback) { FSE_emergencyDistrib(normalizedCounter, nbSymbols, points); return; }   // Fallback mode
            if (reduction < 1) reduction=1;
            if (reduction >= normalizedCounter[rank[3]]) reduction=normalizedCounter[rank[3]]-1;
            fallback = (reduction==0);
            normalizedCounter[rank[3]]-=reduction;
            points-=reduction;
            s=0;
        }
    }
}

// New faster version
int FSE_normalizeCount (short* normalizedCounter, int tableLog, unsigned int* count, int total, int nbSymbols)
{
    // Check
    if (tableLog==0) tableLog = FSE_DEFAULT_TABLELOG;
    if ((FSE_highbit(total-1)-2) < tableLog) tableLog = FSE_highbit(total-1)-2;   // Useless accuracy
    if ((FSE_highbit(nbSymbols)+1) > tableLog) tableLog = FSE_highbit(nbSymbols)+1;   // Need a minimum to represent all symbol values
    if (tableLog < FSE_MIN_TABLELOG) tableLog = FSE_MIN_TABLELOG;
    if (tableLog > FSE_MAX_TABLELOG) return -1;   // Unsupported size

    {
        U32 const rtbTable[] = {     0, 473195, 504333, 520860, 550000, 700000, 750000, 830000 };
        U64 const scale = 62 - tableLog;
        U64 const step = ((U64)1<<62) / total;   // <== (lone) division detected...
        U64 const vStep = 1ULL<<(scale-20);
        int stillToDistribute = 1<<tableLog;
        short s;
        short largest=0, largestP=0;
        U32 lowThreshold = total >> tableLog;

        for (s=0; s<nbSymbols; s++)
        {
            if (count[s] == (U32) total) return 0;   // There is only one symbol
            if (count[s]==0) { normalizedCounter[s]=0; continue; }
            if (count[s] <= lowThreshold)
            {
                normalizedCounter[s] = -1;
                stillToDistribute--;
            }
            else
            {
                short proba = (short)((count[s]*step) >> scale);
                U64 restToBeat;
                if (proba<8)
                {
                    restToBeat = vStep * rtbTable[proba];
                    proba += (count[s]*step) - ((U64)proba<<scale) > restToBeat;
                }
                if (proba > largestP) { largestP=proba; largest=s; }
                normalizedCounter[s] = proba;
                stillToDistribute -= proba;
            }
        }
        if ((int)normalizedCounter[largest] <= -stillToDistribute+8)   // largest cant accomodate that amount
            FSE_distribNpts(normalizedCounter, nbSymbols, (short)(-stillToDistribute));   // Fallback
        else normalizedCounter[largest] += (short)stillToDistribute;
    }

    /*
    {   // Print Table
        int s;
        for (s=0; s<nbSymbols; s++)
            printf("%3i: %4i \n", s, normalizedCounter[s]);
        getchar();
    }
    */

    return tableLog;
}


void* FSE_initCompressionStream(void** op)
{
    void* start = *op;
    *((BYTE**)op) += 4;   // Space to write end of bitStream
    return start;
}


void FSE_initStateAndPtrs(ptrdiff_t* state, const void** symbolTT, const void** stateTable, const void* CTable)
{
    const int tableLog = ( (U16*) CTable) [0];
    *state = (ptrdiff_t)1<<tableLog;
    *stateTable = (void*)(((const U16*) CTable) + 2);
    *symbolTT = (void*)(((const U16*)(*stateTable)) + ((ptrdiff_t)1<<tableLog));
}


void FSE_encodeByte(ptrdiff_t* state, bitStream_forward_t* bitC, BYTE symbol, const void* CTablePtr1, const void* CTablePtr2)
{
    const FSE_symbolCompressionTransform* const symbolTT = (const FSE_symbolCompressionTransform*) CTablePtr1;
    const U16* const stateTable = (const U16*) CTablePtr2;
    int nbBitsOut  = symbolTT[symbol].minBitsOut;
    nbBitsOut -= (int)((symbolTT[symbol].maxState - *state) >> 31);
    FSE_addBits(bitC, *state, nbBitsOut);
    *state = stateTable[ (*state>>nbBitsOut) + symbolTT[symbol].deltaFindState];
}


int FSE_closeCompressionStream(void* outPtr, bitStream_forward_t* bitC, void* compressionStreamDescriptor, int id)
{
    BYTE* p;
    U32 descriptor;

    FSE_flushBits(&outPtr, bitC);

    p = (BYTE*)outPtr; p += bitC->bitPos > 0;
    bitC->bitPos = 8 - bitC->bitPos; if (bitC->bitPos==8) bitC->bitPos=0;

    descriptor = (U32)(p - (BYTE*)compressionStreamDescriptor) << 3;
    descriptor += bitC->bitPos;
    descriptor += (id-1)<<30;   // optional field [1-4]
    *(U32*)compressionStreamDescriptor = descriptor;

    return (int)(p-(BYTE*)compressionStreamDescriptor);
}


int FSE_flushStates(void** outPtr, bitStream_forward_t* bitC,
                    int nbStates, ptrdiff_t state1, ptrdiff_t state2, const void* CTable)
{
    const int tableLog = ( (U16*) CTable) [0];

    if ((nbStates > 2) || (nbStates < 1)) return -1;

    if (nbStates==2) { FSE_addBits(bitC, state2, tableLog); FSE_flushBits(outPtr, bitC); }
    FSE_addBits(bitC, state1, tableLog); FSE_flushBits(outPtr, bitC);

    return 0;
}


FORCE_INLINE int FSE_compress_usingCTable_generic (void* dest, const unsigned char* source, int sourceSize, const void* CTable, int ilp)
{
    const BYTE* const istart = (const BYTE*) source;
    const BYTE* ip;
    const BYTE* const iend = istart + sourceSize;

    BYTE* op = (BYTE*) dest;
    int nbStreams = 1 + ilp;
    U32* streamSizePtr;
    ptrdiff_t state1;
    ptrdiff_t state2;
    bitStream_forward_t bitC = {0,0};   // According to C90/C99, {0} should be enough. Nonetheless, GCC complains....
    const void* stateTable;
    const void* symbolTT;


    streamSizePtr = (U32*)FSE_initCompressionStream((void**)&op);
    FSE_initStateAndPtrs(&state1, &symbolTT, &stateTable, CTable);
    state2 = state1;

    ip=iend;

    // join to even
    if (sourceSize & 1)
    {
        FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);
        FSE_flushBits((void**)&op, &bitC);
    }

    // join to mod 4 (if necessary)
    if ((sizeof(size_t)*8 > FSE_MAX_TABLELOG*4+7 ) && (sourceSize & 2))   // test bit 2
    {
        FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);
        if (ilp) FSE_encodeByte(&state2, &bitC, *--ip, symbolTT, stateTable);
        else FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);
        FSE_flushBits((void**)&op, &bitC);
    }

    // 2 or 4 per loop
    while (ip>istart)
    {
        FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);

        if (sizeof(size_t)*8 < FSE_MAX_TABLELOG*2+7 )   // this test needs to be static
            FSE_flushBits((void**)&op, &bitC);

        if (ilp) FSE_encodeByte(&state2, &bitC, *--ip, symbolTT, stateTable);
        else FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);

        if (sizeof(size_t)*8 > FSE_MAX_TABLELOG*4+7 )   // this test needs to be static
        {
            FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);

            if (ilp) FSE_encodeByte(&state2, &bitC, *--ip, symbolTT, stateTable);
            else FSE_encodeByte(&state1, &bitC, *--ip, symbolTT, stateTable);
        }

        FSE_flushBits((void**)&op, &bitC);
    }

    FSE_flushStates((void**)&op, &bitC, nbStreams, state1, state2, CTable);
    return FSE_closeCompressionStream(op, &bitC, streamSizePtr, nbStreams);
}


int FSE_compress_usingCTable (void* dest, const unsigned char* source, int sourceSize, const void* CTable)
{
    return FSE_compress_usingCTable_generic(dest, source, sourceSize, CTable, FSE_ILP);
}


int FSE_writeSingleChar (BYTE *out, BYTE symbol)
{
    *out++=1;     // Header means ==> 1 symbol repeated across the whole sequence
    *out=symbol;
    return 2;
}

int FSE_noCompression (BYTE* out, const BYTE* in, int isize)
{
    *out++=0;     // Header means ==> uncompressed
    memcpy (out, in, isize);
    return (isize+1);
}


int FSE_compress2 (void* dest, const unsigned char* source, int sourceSize, int nbSymbols, int tableLog)
{
    const BYTE* const istart = (const BYTE*) source;
    const BYTE* ip = istart;

    BYTE* const ostart = (BYTE*) dest;
    BYTE* op = ostart;

    U32   count[FSE_MAX_NB_SYMBOLS];
    S16   norm[FSE_MAX_NB_SYMBOLS];
    CTable_max_t CTable;
    int errorCode;

    // early out
    if (sourceSize <= 1) return FSE_noCompression (ostart, istart, sourceSize);
    if (!nbSymbols) nbSymbols = FSE_MAX_NB_SYMBOLS;
    if (!tableLog) tableLog = FSE_DEFAULT_TABLELOG;

    // Scan input and build symbol stats
    errorCode = FSE_count (count, ip, sourceSize, nbSymbols);
    if (errorCode==-1) return -1;
    if (errorCode==1) return FSE_writeSingleChar (ostart, *istart);   // Only 0 is present
    nbSymbols = errorCode;

    errorCode = FSE_normalizeCount (norm, tableLog, count, sourceSize, nbSymbols);
    if (errorCode==-1) return -1;
    if (errorCode==0) return FSE_writeSingleChar (ostart, *istart);
    tableLog = errorCode;

    // Write table description header
    errorCode = FSE_writeHeader (op, norm, nbSymbols, tableLog);
    if (errorCode==-1) return -1;
    op += errorCode;

    // Compress
    errorCode = FSE_buildCTable (&CTable, norm, nbSymbols, tableLog);
    if (errorCode==-1) return -1;
    op += FSE_compress_usingCTable (op, ip, sourceSize, &CTable);

    // check compressibility
    if ( (op-ostart) >= (sourceSize-1) ) return FSE_noCompression (ostart, istart, sourceSize);

    return (int) (op-ostart);
}


int FSE_compress (void* dest, const unsigned char* source, int sourceSize)
{
    return FSE_compress2(dest, source, sourceSize, FSE_MAX_NB_SYMBOLS, FSE_DEFAULT_TABLELOG);
}


/*********************************************************
   Decompression (Byte symbols)
*********************************************************/
int FSE_decompressRaw (void* out, int osize, const BYTE* in)
{
    memcpy (out, in+1, osize);
    return osize+1;
}

int FSE_decompressSingleSymbol (void* out, int osize, const BYTE symbol)
{
    memset (out, symbol, osize);
    return 2;
}


void FSE_updateBitStream(bitStream_backward_t* bitC, const void** ip)
{
    *((BYTE**)ip) -= bitC->bitsConsumed >> 3;
    bitC->bitContainer = * (U32*) (*ip);
    bitC->bitsConsumed &= 7;
}


FORCE_INLINE const void* FSE_initDecompressionStream_generic(
    const void** p, bitStream_backward_t* bitC, int* optionalId,
    int maxCompressedSize, int safe)
{
    const BYTE* iend;
    const BYTE* ip = (const BYTE*)*p;
    U32 descriptor;

    descriptor = * (U32*) ip;
    *optionalId = (descriptor >> 30) + 1;
    descriptor &= 0x3FFFFFFF;
    bitC->bitsConsumed = descriptor & 7;
    descriptor >>= 3;

    iend = ip + descriptor;
    if (safe) if (iend > ip+maxCompressedSize) return NULL;
    ip = iend - 4;
    *p = (const void*)ip;
    FSE_updateBitStream(bitC, p);

    return (void*)iend;
}

const void* FSE_initDecompressionStream (const void** p, bitStream_backward_t* bitC, int* optionalId)
{ return FSE_initDecompressionStream_generic(p, bitC, optionalId, 0, 0); }

const void* FSE_initDecompressionStream_safe (const void** p, bitStream_backward_t* bitC, int* optionalId, int maxCompressedSize)
{ return FSE_initDecompressionStream_generic(p, bitC, optionalId, maxCompressedSize, 1); }


void FSE_initDStates(
    int nbStates, unsigned int* state1, unsigned int* state2,
    const void** p, bitStream_backward_t* bitC,
    const int tableLog)
{
    *state1 = FSE_readBits(bitC, tableLog); FSE_updateBitStream(bitC, p);
    if (nbStates>=2) { *state2 = FSE_readBits(bitC, tableLog); FSE_updateBitStream(bitC, p); }
}


unsigned int FSE_readBits(bitStream_backward_t* bitC, unsigned int nbBits)
{
    U32 value = ((bitC->bitContainer << bitC->bitsConsumed) >> 1) >> (31-nbBits);
    bitC->bitsConsumed += nbBits;
    return value;
}


BYTE FSE_decodeSymbol(U32* state, bitStream_backward_t* bitC, const void* DTable)
{
    const FSE_decode_t* const decodeTable = (const FSE_decode_t*) DTable;
    BYTE symbol;
    U32 lowBits;
    const U32 nbBits = decodeTable[*state].nbBits;

    symbol = decodeTable[*state].symbol;
    lowBits = FSE_readBits(bitC, nbBits);
    *state = decodeTable[*state].newState + lowBits;

    return symbol;
}


int FSE_closeDecompressionStream(const void* decompressionStreamDescriptor, const void* input)
{ return (int)((BYTE*)decompressionStreamDescriptor - (BYTE*)input); }


FORCE_INLINE int FSE_decompressStreams_usingDTable_generic(
    unsigned char* dest, const int originalSize, const void* compressed, int maxCompressedSize,
    const void* DTable, const int tableLog, int safe, int nbStates)
{
    const void* ip = compressed;
    const void* iend;
    BYTE* op = (BYTE*) dest;
    BYTE* const oend = op + originalSize;
    BYTE* const olimit = oend-1;
    bitStream_backward_t bitC;
    U32 state1;
    U32 state2;

    // Init
    if (safe) iend = FSE_initDecompressionStream_safe(&ip, &bitC, &nbStates, maxCompressedSize);
    else iend = FSE_initDecompressionStream(&ip, &bitC, &nbStates);
    if (iend==NULL) return -1;
    FSE_initDStates(nbStates, &state1, &state2, &ip, &bitC, tableLog);

    // 2 symbols per loop
    while( ((safe) && ((op<olimit) && (ip>=compressed)))
        || ((!safe) && (op<olimit)) )
    {
        if (nbStates==2) *op++ = FSE_decodeSymbol(&state2, &bitC, DTable);
        else *op++ = FSE_decodeSymbol(&state1, &bitC, DTable);
        if (FSE_MAX_TABLELOG*2+7 > sizeof(U32)*8) FSE_updateBitStream(&bitC, &ip);  // Need this test to be static
        *op++ = FSE_decodeSymbol(&state1, &bitC, DTable);
        FSE_updateBitStream(&bitC, &ip);
    }

    // last symbol
    if ( ((safe) && ((op<oend) && (ip>=compressed)))
        || ((!safe) && (op<oend)) )
    {
        *op++ = FSE_decodeSymbol(&state1, &bitC, DTable);
        FSE_updateBitStream(&bitC, &ip);
    }

    if ((ip!=compressed) || bitC.bitsConsumed) return -1;   // Not fully decoded stream

    return FSE_closeDecompressionStream(iend, ip);
}

U32 FSE_getNbStates(const void* buffer)
{
    U32 descriptor = * (U32*)buffer;
    return (descriptor>>30) + 1;
}

FORCE_INLINE int FSE_decompress_usingDTable_generic(
    unsigned char* dest, const int originalSize, const void* compressed, int maxCompressedSize,
    const void* DTable, const int tableLog, int safe)
{
    U32 nbStates = FSE_getNbStates(compressed);
    if (nbStates==2)
        return FSE_decompressStreams_usingDTable_generic(dest, originalSize, compressed, maxCompressedSize, DTable, tableLog, safe, 2);
    if (nbStates==1)
        return FSE_decompressStreams_usingDTable_generic(dest, originalSize, compressed, maxCompressedSize, DTable, tableLog, safe, 1);
    return -1;   // should not happen
}

int FSE_decompress_usingDTable (unsigned char* dest, const int originalSize, const void* compressed, const void* DTable, const int tableLog)
{ return FSE_decompress_usingDTable_generic(dest, originalSize, compressed, 0, DTable, tableLog, 0); }

int FSE_decompress_usingDTable_safe (unsigned char* dest, const int originalSize, const void* compressed, int maxCompressedSize, const void* DTable, const int tableLog)
{ return FSE_decompress_usingDTable_generic(dest, originalSize, compressed, maxCompressedSize, DTable, tableLog, 1); }


FORCE_INLINE int FSE_decompress_generic (
    unsigned char* dest, int originalSize,
    const void* compressed, int maxCompressedSize, int safe)
{
    const BYTE* const istart = (const BYTE*)compressed;
    const BYTE* ip = istart;
    short   counting[FSE_MAX_NB_SYMBOLS];
    FSE_decode_t DTable[FSE_MAX_TABLESIZE];
    BYTE  headerId;
    int nbSymbols;
    int tableLog;
    int errorCode;

    // headerId early outs
    if ((safe) && (maxCompressedSize<2)) return -1;   // too small input size
    headerId = ip[0] & 3;
    if (ip[0]==0) return FSE_decompressRaw (dest, originalSize, istart);
    if (ip[0]==1) return FSE_decompressSingleSymbol (dest, originalSize, istart[1]);
    if (headerId!=2) return -1;   // unused headerId

    // normal FSE decoding mode
    errorCode = FSE_readHeader (counting, &nbSymbols, &tableLog, istart);
    if (errorCode==-1) return -1;
    ip += errorCode;

    errorCode = FSE_buildDTable (DTable, counting, nbSymbols, tableLog);
    if (errorCode==-1) return -1;

    if (safe) errorCode = FSE_decompress_usingDTable_safe (dest, originalSize, ip, maxCompressedSize, DTable, tableLog);
    else errorCode = FSE_decompress_usingDTable (dest, originalSize, ip, DTable, tableLog);
    if (errorCode==-1) return -1;
    ip += errorCode;

    return (int) (ip-istart);
}

int FSE_decompress (unsigned char* dest, int originalSize, const void* compressed)
{ return FSE_decompress_generic(dest, originalSize, compressed, 0, 0); }

int FSE_decompress_safe (unsigned char* dest, int originalSize, const void* compressed, int maxCompressedSize)
{ return FSE_decompress_generic(dest, originalSize, compressed, maxCompressedSize, 1); }


#endif   // FSE_DONTINCLUDECORE

/*
  2nd part of the file
  designed to be included
  for type-specific functions (template equivalent in C)
  Objective is to write such functions only once, for better maintenance
*/

// checks
#ifndef FSE_FUNCTION_EXTENSION
#  error "FSE_FUNCTION_EXTENSION must be defined"
#endif
#ifndef FSE_FUNCTION_TYPE
#  error "FSE_FUNCTION_TYPE must be defined"
#endif

// Function names
#define FSE_CAT(X,Y) X##Y
#define FSE_FUNCTION_NAME(X,Y) FSE_CAT(X,Y)
#define FSE_TYPE_NAME(X,Y) FSE_CAT(X,Y)


// Functions
int FSE_FUNCTION_NAME(FSE_count, FSE_FUNCTION_EXTENSION) (unsigned int* count, const FSE_FUNCTION_TYPE* source, int sourceSize, int maxNbSymbols)
{
    const FSE_FUNCTION_TYPE* ip = (const FSE_FUNCTION_TYPE*) source;
    const FSE_FUNCTION_TYPE* const iend = ip+sourceSize;
    int   i;

    U32   Counting1[FSE_MAX_NB_SYMBOLS] = {0};
    U32   Counting2[FSE_MAX_NB_SYMBOLS] = {0};
    U32   Counting3[FSE_MAX_NB_SYMBOLS] = {0};
    U32   Counting4[FSE_MAX_NB_SYMBOLS] = {0};

    // Init checks
    if (maxNbSymbols > FSE_MAX_NB_SYMBOLS) return -1;        // maxNbSymbols too large : unsupported
    if (!maxNbSymbols) maxNbSymbols = FSE_MAX_NB_SYMBOLS;    // 0: default
    if (!sourceSize) return -1;                              // Error : no input

    while (ip < iend-3)
    {
        Counting1[*ip++]++;
        Counting2[*ip++]++;
        Counting3[*ip++]++;
        Counting4[*ip++]++;
    }
    while (ip<iend) Counting1[*ip++]++;

    for (i=0; i<maxNbSymbols; i++) count[i] = Counting1[i] + Counting2[i] + Counting3[i] + Counting4[i];

    while (!count[maxNbSymbols-1]) maxNbSymbols--;
    return maxNbSymbols;
}


#define FSE_TABLESTEP(tableSize) ((tableSize>>1) + (tableSize>>3) + 3)

int FSE_FUNCTION_NAME(FSE_buildCTable, FSE_FUNCTION_EXTENSION)
(void* CTable, const short* normalizedCounter, int nbSymbols, int tableLog)
{
    const int tableSize = 1 << tableLog;
    const int tableMask = tableSize - 1;
    U16* tableU16 = ( (U16*) CTable) + 2;
    FSE_symbolCompressionTransform* symbolTT = (FSE_symbolCompressionTransform*) (tableU16 + tableSize);
    const int step = FSE_TABLESTEP(tableSize);
    int cumul[FSE_MAX_NB_SYMBOLS+1];
    U32 position = 0;
    FSE_FUNCTION_TYPE tableSymbol[FSE_MAX_TABLESIZE];
    U32 highThreshold = tableSize-1;
    int s;
    int i;

    // header
    tableU16[-2] = (U16) tableLog;
    tableU16[-1] = (U16) nbSymbols;

    // For explanations on how to distribute symbol values over the table :
    // http://fastcompression.blogspot.fr/2014/02/fse-distributing-symbol-values.html

    // symbol start positions
    cumul[0] = 0;
    for (i=1; i<=nbSymbols; i++)
    {
        if (normalizedCounter[i-1]==-1)   // Low prob symbol
        {
            cumul[i] = cumul[i-1] + 1;
            tableSymbol[highThreshold--] = (FSE_FUNCTION_TYPE)(i-1);
        }
        else
        cumul[i] = cumul[i-1] + normalizedCounter[i-1];
    }
    cumul[nbSymbols] = tableSize+1;

    // Spread symbols
    for (s=0; s<nbSymbols; s++)
    {
        int i;
        for (i=0; i<normalizedCounter[s]; i++)
        {
            tableSymbol[position] = (FSE_FUNCTION_TYPE)s;
            position = (position + step) & tableMask;
            while (position > highThreshold) position = (position + step) & tableMask;   // Lowprob area
        }
    }

    if (position!=0) return -1;   // Must have gone through all positions, otherwise normalizedCount is not correct

    // Build table
    for (i=0; i<tableSize; i++)
    {
        FSE_FUNCTION_TYPE s = tableSymbol[i];
        tableU16[cumul[s]++] = (U16) (tableSize+i);
    }

    // Build Symbol Transformation Table
    {
        int s;
        int total = 0;
        for (s=0; s<nbSymbols; s++)
        {
            switch (normalizedCounter[s])
            {
            case 0:
                break;
            case -1:
            case 1:
                symbolTT[s].minBitsOut = (BYTE) tableLog;
                symbolTT[s].deltaFindState = total - 1;
                total ++;
                symbolTT[s].maxState = (U16) ( (tableSize*2) - 1);   // ensures state <= maxState
                break;
            default :
                symbolTT[s].minBitsOut = (BYTE) ( (tableLog-1) - FSE_highbit (normalizedCounter[s]-1) );
                symbolTT[s].deltaFindState = total - normalizedCounter[s];
                total +=  normalizedCounter[s];
                symbolTT[s].maxState = (U16) ( (normalizedCounter[s]<< (symbolTT[s].minBitsOut+1) ) - 1);
            }
        }
    }

    return 0;
}


#define FSE_DECODE_TYPE FSE_TYPE_NAME(FSE_decode_t, FSE_FUNCTION_EXTENSION)

int FSE_FUNCTION_NAME(FSE_sizeof_DTable, FSE_FUNCTION_EXTENSION) (int tableLog)
{ return (int) ( (1<<tableLog) * (int) sizeof (FSE_DECODE_TYPE) ); }

int FSE_FUNCTION_NAME(FSE_buildDTable, FSE_FUNCTION_EXTENSION)
(void* DTable, const short* const normalizedCounter, int nbSymbols, int tableLog)
{
    FSE_DECODE_TYPE* const tableDecode = (FSE_DECODE_TYPE*) DTable;
    const U32 tableSize = 1 << tableLog;
    const U32 tableMask = tableSize-1;
    const U32 step = FSE_TABLESTEP(tableSize);
    U16 symbolNext[FSE_MAX_NB_SYMBOLS];
    U32 position = 0;
    U32 highThreshold = tableSize-1;
    int s;

    // Checks
    if (nbSymbols > FSE_MAX_NB_SYMBOLS) return -1;
    if (tableLog > FSE_MAX_TABLELOG) return -1;

    // Low prob symbols
    for (s=0; s<nbSymbols; s++)
        if (normalizedCounter[s]==-1)
            tableDecode[highThreshold--].symbol = (FSE_FUNCTION_TYPE)s;

    // Spread symbols
    for (s=0; s<nbSymbols; s++)
    {
        int i;
        for (i=0; i<normalizedCounter[s]; i++)
        {
            tableDecode[position].symbol = (FSE_FUNCTION_TYPE)s;
            position = (position + step) & tableMask;
            while (position > highThreshold) position = (position + step) & tableMask;   // lowprob area
        }
    }

    if (position!=0) return -1;   // position must use all positions, otherwise normalizedCounter is incorrect

    // Calculate symbol next
    for (s=0; s<nbSymbols; s++) symbolNext[s] = FSE_abs(normalizedCounter[s]);

    // Build table Decoding table
    {
        U32 i;
        for (i=0; i<tableSize; i++)
        {
            FSE_FUNCTION_TYPE s = tableDecode[i].symbol;
            U16 nextState = symbolNext[s]++;
            tableDecode[i].nbBits = (BYTE) (tableLog - FSE_highbit (nextState) );
            tableDecode[i].newState = (U16) ( (nextState << tableDecode[i].nbBits) - tableSize);
        }
    }

    return 0;
}
